@@ -33,6 +33,7 @@ import base64
 import operator
 import asyncio
 import inspect
+from collections import defaultdict
 from functools import wraps, partial
 from itertools import repeat
 from decimal import Decimal
@@ -54,7 +55,7 @@ from .address_synchronizer import TX_HEIGHT_LOCAL
 from .mnemonic import Mnemonic
 from .lnutil import SENT, RECEIVED
 from .lnutil import LnFeatures
-from .lnutil import ln_dummy_address
+from .lnutil import extract_nodeid
 from .lnpeer import channel_id_from_funding_tx
 from .plugin import run_hook
 from .version import ELECTRUM_VERSION
@@ -75,9 +76,12 @@ class NotSynchronizedException(Exception):
     pass
 
 
+def satoshis_or_max(amount):
+    return satoshis(amount) if amount != '!' else '!'
+
 def satoshis(amount):
     # satoshi conversion must not be performed by the parser
-    return int(COIN*Decimal(amount)) if amount not in ['!', None] else amount
+    return int(COIN*Decimal(amount)) if amount is not None else None
 
 def format_satoshis(x):
     return str(Decimal(x)/COIN) if x is not None else None
@@ -210,7 +214,8 @@ class Commands:
     @command('n')
     async def stop(self):
         """Stop daemon"""
-        self.daemon.stop()
+        # TODO it would be nice if this could stop the GUI too
+        await self.daemon.stop()
         return "Daemon stopped"
 
     @command('n')
@@ -231,7 +236,7 @@ class Commands:
     @command('n')
     async def close_wallet(self, wallet_path=None):
         """Close wallet"""
-        return self.daemon.stop_wallet(wallet_path)
+        return await self.daemon._stop_wallet(wallet_path)
 
     @command('')
     async def create(self, passphrase=None, password=None, encrypt_file=True, seed_type=None, wallet_path=None):
@@ -293,6 +298,7 @@ class Commands:
     def _setconfig_normalize_value(cls, key, value):
         if key not in ('rpcuser', 'rpcpassword'):
             value = json_decode(value)
+            # call literal_eval for backward compatibility (see #4225)
             try:
                 value = ast.literal_eval(value)
             except:
@@ -303,6 +309,10 @@ class Commands:
     async def setconfig(self, key, value):
         """Set a configuration variable. 'value' may be a string or a Python expression."""
         value = self._setconfig_normalize_value(key, value)
+        if self.daemon and key == 'rpcuser':
+            self.daemon.commands_server.rpc_user = value
+        if self.daemon and key == 'rpcpassword':
+            self.daemon.commands_server.rpc_password = value
         self.config.set_key(key, value)
         return True
 
@@ -313,10 +323,10 @@ class Commands:
         return self.config.get_ssl_domain()
 
     @command('')
-    async def make_seed(self, nbits=132, language=None, seed_type=None):
+    async def make_seed(self, nbits=None, language=None, seed_type=None):
         """Create a seed"""
         from .mnemonic import Mnemonic
-        s = Mnemonic(language).make_seed(seed_type, num_bits=nbits)
+        s = Mnemonic(language).make_seed(seed_type=seed_type, num_bits=nbits)
         return s
 
     @command('n')
@@ -355,7 +365,7 @@ class Commands:
         """
         keypairs = {}
         inputs = []  # type: List[PartialTxInput]
-        locktime = jsontx.get('lockTime', 0)
+        locktime = jsontx.get('locktime', 0)
         for txin_dict in jsontx.get('inputs'):
             if txin_dict.get('prevout_hash') is not None and txin_dict.get('prevout_n') is not None:
                 prevout = TxOutpoint(txid=bfh(txin_dict['prevout_hash']), out_idx=int(txin_dict['prevout_n']))
@@ -364,7 +374,7 @@ class Commands:
             else:
                 raise Exception("missing prevout for txin")
             txin = PartialTxInput(prevout=prevout)
-            txin._trusted_value_sats = int(txin_dict['value'])
+            txin._trusted_value_sats = int(txin_dict.get('value', txin_dict['value_sats']))
             nsequence = txin_dict.get('nsequence', None)
             if nsequence is not None:
                 txin.nsequence = nsequence
@@ -378,22 +388,41 @@ class Commands:
                 txin.num_sig = 1
             inputs.append(txin)
 
-        outputs = [PartialTxOutput.from_address_and_value(txout['address'], int(txout['value']))
+        outputs = [PartialTxOutput.from_address_and_value(txout['address'], int(txout.get('value', txout['value_sats'])))
                    for txout in jsontx.get('outputs')]
         tx = PartialTransaction.from_io(inputs, outputs, locktime=locktime)
         tx.sign(keypairs)
         return tx.serialize()
 
-    @command('wp')
-    async def signtransaction(self, tx, privkey=None, password=None, wallet: Abstract_Wallet = None):
-        """Sign a transaction. The wallet keys will be used unless a private key is provided."""
+    @command('')
+    async def signtransaction_with_privkey(self, tx, privkey):
+        """Sign a transaction. The provided list of private keys will be used to sign the transaction."""
         tx = tx_from_any(tx)
-        if privkey:
-            txin_type, privkey2, compressed = bitcoin.deserialize_privkey(privkey)
-            pubkey = ecc.ECPrivkey(privkey2).get_public_key_bytes(compressed=compressed).hex()
-            tx.sign({pubkey:(privkey2, compressed)})
-        else:
-            wallet.sign_transaction(tx, password)
+
+        txins_dict = defaultdict(list)
+        for txin in tx.inputs():
+            txins_dict[txin.address].append(txin)
+
+        if not isinstance(privkey, list):
+            privkey = [privkey]
+
+        for priv in privkey:
+            txin_type, priv2, compressed = bitcoin.deserialize_privkey(priv)
+            pubkey = ecc.ECPrivkey(priv2).get_public_key_bytes(compressed=compressed)
+            address = bitcoin.pubkey_to_address(txin_type, pubkey.hex())
+            if address in txins_dict.keys():
+                for txin in txins_dict[address]:
+                    txin.pubkeys = [pubkey]
+                    txin.script_type = txin_type
+                tx.sign({pubkey.hex(): (priv2, compressed)})
+
+        return tx.serialize()
+
+    @command('wp')
+    async def signtransaction(self, tx, password=None, wallet: Abstract_Wallet = None):
+        """Sign a transaction. The wallet keys will be used to sign the transaction."""
+        tx = tx_from_any(tx)
+        wallet.sign_transaction(tx, password)
         return tx.serialize()
 
     @command('')
@@ -418,15 +447,27 @@ class Commands:
         return {'address':address, 'redeemScript':redeem_script}
 
     @command('w')
-    async def freeze(self, address, wallet: Abstract_Wallet = None):
+    async def freeze(self, address: str, wallet: Abstract_Wallet = None):
         """Freeze address. Freeze the funds at one of your wallet\'s addresses"""
         return wallet.set_frozen_state_of_addresses([address], True)
 
     @command('w')
-    async def unfreeze(self, address, wallet: Abstract_Wallet = None):
+    async def unfreeze(self, address: str, wallet: Abstract_Wallet = None):
         """Unfreeze address. Unfreeze the funds at one of your wallet\'s address"""
         return wallet.set_frozen_state_of_addresses([address], False)
 
+    @command('w')
+    async def freeze_utxo(self, coin: str, wallet: Abstract_Wallet = None):
+        """Freeze a UTXO so that the wallet will not spend it."""
+        wallet.set_frozen_state_of_coins([coin], True)
+        return True
+
+    @command('w')
+    async def unfreeze_utxo(self, coin: str, wallet: Abstract_Wallet = None):
+        """Unfreeze a UTXO so that the wallet might spend it."""
+        wallet.set_frozen_state_of_coins([coin], False)
+        return True
+
     @command('wp')
     async def getprivatekeys(self, address, password=None, wallet: Abstract_Wallet = None):
         """Get private keys of addresses. You may pass a single wallet address, or a list of wallet addresses."""
@@ -561,12 +602,14 @@ class Commands:
         privkeys = privkey.split()
         self.nocheck = nocheck
         #dest = self._resolver(destination)
-        tx = sweep(privkeys,
-                   network=self.network,
-                   config=self.config,
-                   to_address=destination,
-                   fee=tx_fee,
-                   imax=imax)
+        tx = await sweep(
+            privkeys,
+            network=self.network,
+            config=self.config,
+            to_address=destination,
+            fee=tx_fee,
+            imax=imax,
+        )
         return tx.serialize() if tx else None
 
     @command('wp')
@@ -593,7 +636,7 @@ class Commands:
         domain_coins = from_coins.split(',') if from_coins else None
         change_addr = self._resolver(change_addr, wallet)
         domain_addr = None if domain_addr is None else map(self._resolver, domain_addr, repeat(wallet))
-        amount_sat = satoshis(amount)
+        amount_sat = satoshis_or_max(amount)
         outputs = [PartialTxOutput.from_address_and_value(destination, amount_sat)]
         tx = wallet.create_transaction(
             outputs,
@@ -624,7 +667,7 @@ class Commands:
         final_outputs = []
         for address, amount in outputs:
             address = self._resolver(address, wallet)
-            amount_sat = satoshis(amount)
+            amount_sat = satoshis_or_max(amount)
             final_outputs.append(PartialTxOutput.from_address_and_value(address, amount_sat))
         tx = wallet.create_transaction(
             final_outputs,
@@ -643,10 +686,13 @@ class Commands:
         return result
 
     @command('w')
-    async def onchain_history(self, year=None, show_addresses=False, show_fiat=False, wallet: Abstract_Wallet = None):
+    async def onchain_history(self, year=None, show_addresses=False, show_fiat=False, wallet: Abstract_Wallet = None,
+                              from_height=None, to_height=None):
         """Wallet onchain history. Returns the transaction history of your wallet."""
         kwargs = {
             'show_addresses': show_addresses,
+            'from_height': from_height,
+            'to_height': to_height,
         }
         if year:
             import time
@@ -658,6 +704,7 @@ class Commands:
             from .exchange_rate import FxThread
             fx = FxThread(self.config, None)
             kwargs['fx'] = fx
+
         return json_normalize(wallet.get_detailed_history(**kwargs))
 
     @command('w')
@@ -781,7 +828,8 @@ class Commands:
             f = None
         out = wallet.get_sorted_requests()
         if f is not None:
-            out = list(filter(lambda x: x.status==f, out))
+            out = [req for req in out
+                   if f == wallet.get_request_status(wallet.get_key_for_receive_request(req))]
         return [wallet.export_request(x) for x in out]
 
     @command('w')
@@ -926,13 +974,10 @@ class Commands:
         if not is_hash256_str(txid):
             raise Exception(f"{repr(txid)} is not a txid")
         height = wallet.get_tx_height(txid).height
-        to_delete = {txid}
         if height != TX_HEIGHT_LOCAL:
             raise Exception(f'Only local transactions can be removed. '
                             f'This tx has height: {height} != {TX_HEIGHT_LOCAL}')
-        to_delete |= wallet.get_depending_transactions(txid)
-        for tx_hash in to_delete:
-            wallet.remove_transaction(tx_hash)
+        wallet.remove_transaction(txid)
         wallet.save_db()
 
     @command('wn')
@@ -975,13 +1020,19 @@ class Commands:
     async def open_channel(self, connection_string, amount, push_amount=0, password=None, wallet: Abstract_Wallet = None):
         funding_sat = satoshis(amount)
         push_sat = satoshis(push_amount)
-        dummy_output = PartialTxOutput.from_address_and_value(ln_dummy_address(), funding_sat)
-        funding_tx = wallet.mktx(outputs = [dummy_output], rbf=False, sign=False, nonlocal_only=True)
-        chan, funding_tx = await wallet.lnworker._open_channel_coroutine(connect_str=connection_string,
-                                                                         funding_tx=funding_tx,
-                                                                         funding_sat=funding_sat,
-                                                                         push_sat=push_sat,
-                                                                         password=password)
+        coins = wallet.get_spendable_coins(None)
+        node_id, rest = extract_nodeid(connection_string)
+        funding_tx = wallet.lnworker.mktx_for_open_channel(
+            coins=coins,
+            funding_sat=funding_sat,
+            node_id=node_id,
+            fee_est=None)
+        chan, funding_tx = await wallet.lnworker._open_channel_coroutine(
+            connect_str=connection_string,
+            funding_tx=funding_tx,
+            funding_sat=funding_sat,
+            push_sat=push_sat,
+            password=password)
         return chan.funding_outpoint.to_str()
 
     @command('')
@@ -995,7 +1046,7 @@ class Commands:
         lnaddr = lnworker._check_invoice(invoice)
         payment_hash = lnaddr.paymenthash
         wallet.save_invoice(LNInvoice.from_bech32(invoice))
-        success, log = await lnworker._pay(invoice, attempts=attempts)
+        success, log = await lnworker.pay_invoice(invoice, attempts=attempts)
         return {
             'payment_hash': payment_hash.hex(),
             'success': success,
@@ -1010,13 +1061,15 @@ class Commands:
 
     @command('w')
     async def list_channels(self, wallet: Abstract_Wallet = None):
-        # we output the funding_outpoint instead of the channel_id because lnd uses channel_point (funding outpoint) to identify channels
+        # FIXME: we need to be online to display capacity of backups
         from .lnutil import LOCAL, REMOTE, format_short_channel_id
-        l = list(wallet.lnworker.channels.items())
+        channels = list(wallet.lnworker.channels.items())
+        backups = list(wallet.lnworker.channel_backups.items())
         return [
             {
+                'type': 'CHANNEL',
                 'short_channel_id': format_short_channel_id(chan.short_channel_id) if chan.short_channel_id else None,
-                'channel_id': bh2u(chan.channel_id),
+                'channel_id': chan.channel_id.hex(),
                 'channel_point': chan.funding_outpoint.to_str(),
                 'state': chan.get_state().name,
                 'peer_state': chan.peer_state.name,
@@ -1027,7 +1080,15 @@ class Commands:
                 'remote_reserve': chan.config[LOCAL].reserve_sat,
                 'local_unsettled_sent': chan.balance_tied_up_in_htlcs_by_direction(LOCAL, direction=SENT) // 1000,
                 'remote_unsettled_sent': chan.balance_tied_up_in_htlcs_by_direction(REMOTE, direction=SENT) // 1000,
-            } for channel_id, chan in l
+            } for channel_id, chan in channels
+        ] + [
+            {
+                'type': 'BACKUP',
+                'short_channel_id': format_short_channel_id(chan.short_channel_id) if chan.short_channel_id else None,
+                'channel_id': chan.channel_id.hex(),
+                'channel_point': chan.funding_outpoint.to_str(),
+                'state': chan.get_state().name,
+            } for channel_id, chan in backups
         ]
 
     @command('wn')
@@ -1036,18 +1097,23 @@ class Commands:
 
     @command('n')
     async def inject_fees(self, fees):
-        import ast
-        self.network.config.fee_estimates = ast.literal_eval(fees)
-        self.network.notify('fee')
+        # e.g. use from Qt console:  inject_fees("{25: 1009, 10: 15962, 5: 18183, 2: 23239}")
+        fee_est = ast.literal_eval(fees)
+        self.network.update_fee_estimates(fee_est=fee_est)
 
     @command('wn')
     async def enable_htlc_settle(self, b: bool, wallet: Abstract_Wallet = None):
-        e = wallet.lnworker.enable_htlc_settle
-        e.set() if b else e.clear()
+        wallet.lnworker.enable_htlc_settle = b
 
     @command('n')
     async def clear_ln_blacklist(self):
-        self.network.path_finder.blacklist.clear()
+        if self.network.path_finder:
+            self.network.path_finder.liquidity_hints.clear_blacklist()
+
+    @command('n')
+    async def reset_liquidity_hints(self):
+        if self.network.path_finder:
+            self.network.path_finder.liquidity_hints.reset_liquidity_hints()
 
     @command('w')
     async def list_invoices(self, wallet: Abstract_Wallet = None):
@@ -1061,6 +1127,17 @@ class Commands:
         coro = wallet.lnworker.force_close_channel(chan_id) if force else wallet.lnworker.close_channel(chan_id)
         return await coro
 
+    @command('wn')
+    async def request_force_close(self, channel_point, connection_string=None, wallet: Abstract_Wallet = None):
+        """
+        Requests the remote to force close a channel.
+        If a connection string is passed, can be used without having state or any backup for the channel.
+        Assumes that channel was originally opened with the same local peer (node_keypair).
+        """
+        txid, index = channel_point.split(':')
+        chan_id, _ = channel_id_from_funding_tx(txid, int(index))
+        await wallet.lnworker.request_force_close(chan_id, connect_str=connection_string)
+
     @command('w')
     async def export_channel_backup(self, channel_point, wallet: Abstract_Wallet = None):
         txid, index = channel_point.split(':')
@@ -1069,7 +1146,7 @@ class Commands:
 
     @command('w')
     async def import_channel_backup(self, encrypted, wallet: Abstract_Wallet = None):
-        return wallet.lnbackups.import_channel_backup(encrypted)
+        return wallet.lnworker.import_channel_backup(encrypted)
 
     @command('wn')
     async def get_channel_ctx(self, channel_point, iknowwhatimdoing=False, wallet: Abstract_Wallet = None):
@@ -1108,7 +1185,11 @@ class Commands:
         else:
             lightning_amount_sat = satoshis(lightning_amount)
             onchain_amount_sat = satoshis(onchain_amount)
-            txid = await wallet.lnworker.swap_manager.normal_swap(lightning_amount_sat, onchain_amount_sat, password)
+            txid = await wallet.lnworker.swap_manager.normal_swap(
+                lightning_amount_sat=lightning_amount_sat,
+                expected_onchain_amount_sat=onchain_amount_sat,
+                password=password,
+            )
         return {
             'txid': txid,
             'lightning_amount': format_satoshis(lightning_amount_sat),
@@ -1132,8 +1213,12 @@ class Commands:
             success = None
         else:
             lightning_amount_sat = satoshis(lightning_amount)
-            onchain_amount_sat = satoshis(onchain_amount)
-            success = await wallet.lnworker.swap_manager.reverse_swap(lightning_amount_sat, onchain_amount_sat)
+            claim_fee = sm.get_claim_fee()
+            onchain_amount_sat = satoshis(onchain_amount) + claim_fee
+            success = await wallet.lnworker.swap_manager.reverse_swap(
+                lightning_amount_sat=lightning_amount_sat,
+                expected_onchain_amount_sat=onchain_amount_sat,
+            )
         return {
             'success': success,
             'lightning_amount': format_satoshis(lightning_amount_sat),
@@ -1217,6 +1302,7 @@ command_options = {
     'to_height':   (None, "Only show transactions that confirmed before given block height"),
     'iknowwhatimdoing': (None, "Acknowledge that I understand the full implications of what I am about to do"),
     'gossip':      (None, "Apply command to gossip node instead of wallet"),
+    'connection_string':      (None, "Lightning network node ID or network address"),
 }
 
 
@@ -1349,6 +1435,7 @@ def get_parser():
     parser_gui.add_argument("-m", action="store_true", dest="hide_gui", default=False, help="hide GUI on startup")
     parser_gui.add_argument("-L", "--lang", dest="language", default=None, help="default language used in GUI")
     parser_gui.add_argument("--daemon", action="store_true", dest="daemon", default=False, help="keep daemon running after GUI is closed")
+    parser_gui.add_argument("--nosegwit", action="store_true", dest="nosegwit", default=False, help="Do not create segwit wallets")
     add_wallet_option(parser_gui)
     add_network_options(parser_gui)
     add_global_options(parser_gui)
